package com.ibm.uia.xml;

/****************************************************************************/
/*                                                                          */
/* Program name: MQ Series Everyplace - Administration Tool                 */
/*                                                                          */
/*  Statement:  Licensed Materials - Property of IBM                        */
/*                                                                          */
/*              MQSeries SupportPac MSE1                                    */
/*              (C) Copyright IBM Corp. 2000                                */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Change history:                                                          */
/*                                                                          */
/* V1.0   15-08-2000  PTvL  Initial release                                 */
/*                                                                          */
/*==========================================================================*/
import java.util.*;
import java.io.*;

/**
 * This type parses a stream of characters into a stream of tags.
 *
 */
public final class TagStream implements XMLConstants {
	private static final String COMMENT_START = "!--";
	
	private int 	_characterCount = -1;
	private int 	_lastCharacterRead = END_OF_FILE;
	private Reader 	_inReader;

	private boolean isEndTag;
	private StringBuffer tagName;	// buffer in which to build the tag name
	private StringBuffer attrName;  // buffer in which to build attribute name
	private StringBuffer attrValue;	// buffer in which to build attribute value
	private Element currentTag;
	
/**
 * TagReader constructor comment.
 */
public TagStream(Reader inReader) {
	super();
	_inReader = inReader;
	attrName = new StringBuffer();
	attrValue = new StringBuffer();
	tagName = new StringBuffer();
}
/* Remove characters up to the end of the comment.
 * This means remove all characters from the internal
 * stream up to and including the next "-->" 
 *
 */
private final void _findEndOfComment() {
	do {
		if (MINUS == _lastCharacterRead) {
			_readCharacter(); // discard '-'
			if (MINUS == _lastCharacterRead) {
				_readCharacter(); // discard '-'
				if (END_OF_TAG == _lastCharacterRead) {
					_readCharacter(); // discard '<'
					return;
				}
			}
		}
		_readCharacter();
	} while (END_OF_FILE != _lastCharacterRead);
}
/* Remove characters up to the end of this tag.
 * This means remove all characters from the internal
 * stream up to and including the next END_OF_TAG 
 * character.  Reply true if a start of tag was found,
 * false otherwise;
 */
private final boolean _findEndOfTag() {
	do {
		if (END_OF_TAG == _lastCharacterRead) {
			_readCharacter();  // discard '<'
			return true;
		} else if (FORWARD_SLASH == _lastCharacterRead) {
			if (!currentTag._isEndTag()) {
				currentTag._setIsEmptyTag(true);
			}
		} else if (START_OF_TAG == _lastCharacterRead) {
			//TODO throw error here: 'end tag without start tag';
		}
		_readCharacter();
	} while (END_OF_FILE != _lastCharacterRead);
	return false;
}
/* Remove characters up to the start of the next tag.
 * This means remove all characters from the internal
 * stream up to and including the next start of tag 
 * character.  Reply true if a start of tag was found,
 * false otherwise;
 *
 * If we find the start of a tag, check to see if this is a comment
 */
private final boolean _findStartOfTag() {
	do {
		if (START_OF_TAG == _lastCharacterRead) {
			_readCharacter();  // discard '<'
			_removeWhitespace();
			_readTagName();
			if (tagName.toString().equals(COMMENT_START)) {
				_readCharacter();  // discard '-'
				_findEndOfComment();
				// now it will continue to look for next tag.....
			} else {
				return true;
			}
		} else if (END_OF_TAG == _lastCharacterRead) {
			//TODO throw error here: 'end tag without start tag';
		}
		_readCharacter();
	} while (END_OF_FILE != _lastCharacterRead);
	return false;
}
private final boolean _isEndOfString() {
	return (END_OF_FILE == _lastCharacterRead 
			|| QUOTE == _lastCharacterRead);
}
private final boolean _isSyntax() {
	return (END_OF_FILE == _lastCharacterRead 
			|| END_OF_TAG == _lastCharacterRead 
			|| FORWARD_SLASH == _lastCharacterRead 
			|| QUOTE == _lastCharacterRead
			|| EQUALS == _lastCharacterRead);
}
/* Parses the stream for the contents of a tag;
 * This process is 'tolerant' in that certain
 * syntactic deformities are currently ignored.
 */
private final Element _parseTagData() {
	currentTag = new Element(_characterCount);
	
	currentTag._setIsEndTag(isEndTag);
	currentTag._setName(tagName.toString());
	if (!currentTag._isEndTag()) {
		while (!_isSyntax()) {
			_removeWhitespace();
			_readTagAttribute();
		}
	} else {
			_removeWhitespace();
	}
	
	if (_findEndOfTag()) {
		currentTag._setEndPos(_characterCount);
		return currentTag;
	}
	
	return null;
}
/* 
 */
private final void _readCharacter() {
	try {
		_lastCharacterRead = _inReader.read();
		_characterCount++;
	} catch (IOException e) {
		_lastCharacterRead = END_OF_FILE;
			//TODO throw error: 'unexpected end of file';
	}	
}
/* Read a tag attribute from the stream.  The atribute name 
 * is assumed to be all the characters upto but not including 
 * the next '='.  If any character other than '=' is detected
 * then the attribute read is aborted;
 * If a '=' was successfully reached then the attribute value 
 * is read from the stream. The attribute value is assumed to be
 * quote delimited.  
 */
private final void _readTagAttribute() {
	int c = -1;
	attrName.setLength(0);
	while (!_isSyntax()) {
		attrName.append((char)_lastCharacterRead);	
		_readCharacter();
	}
	
	if (EQUALS == _lastCharacterRead) {
		attrValue.setLength(0);
		_readCharacter();   // remove EQUALS
		_removeWhitespace();
		if (QUOTE == _lastCharacterRead) {
			_readCharacter();   // remove QUOTE
			while (!_isEndOfString()) {
				attrValue.append((char)_lastCharacterRead);	
				_readCharacter();
			}
			if (QUOTE == _lastCharacterRead) {
				currentTag.setAttribute(attrName.toString().trim(), 
									attrValue.toString().trim());
				_readCharacter();
			}
		}
	}
}
/* Read the name from the stream.
 * All characters upto but not including the next space or
 * syntax mark are part of the tag name
 */
private final void _readTagName() {
	tagName.setLength(0);
	if (FORWARD_SLASH == _lastCharacterRead) {
		//currentTag._setIsEndTag(true);
		isEndTag = true;
		_readCharacter();
	} else {
		isEndTag = false;
	}
	
	while (!_isSyntax() && SPACE != _lastCharacterRead) {
		tagName.append((char)_lastCharacterRead);
		// short circuit on detecting a comment
		if (tagName.length() == 3 && tagName.toString().equals(COMMENT_START)) {
			return;
		}
		_readCharacter();
	}
	//currentTag._setName(tagName.toString());
}
/* Remove whitespace characters
 */
private final void _removeWhitespace() {
	while (SPACE == _lastCharacterRead) {
		_readCharacter();
	};
}
public final Element nextTag() {
	Element tag = null; 
	if (_findStartOfTag()) {
		tag = _parseTagData();
	}
	return tag;
}
}
